library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--library work;
--use work.adcmv3_components.all;

-- BUG: does alway set bit 0 of address byte to zero !!!!
-- REMARK: this is not a bug, but a feature....

entity i2c_slim is
port(
  CLOCK           : in  std_logic;
  RESET           : in  std_logic;
  -- I2C command / setup
  I2C_GO_IN       : in  std_logic; -- startbit to trigger I2C actions
  ACTION_IN       : in  std_logic; -- '0' -> write, '1' -> read
  WORD_IN         : in  std_logic; -- '0' -> byte, '1' -> word
  I2C_SPEED_IN    : in  std_logic_vector(5 downto 0); -- speed adjustment (to be defined)
  I2C_ADDR_IN     : in  std_logic_vector(7 downto 0); -- I2C address byte (R/W bit is ignored)
  I2C_CMD_IN      : in  std_logic_vector(7 downto 0); -- I2C command byte (sent after address byte)
  I2C_DW_IN       : in  std_logic_vector(15 downto 0); -- data word for write command
  I2C_DR_OUT      : out std_logic_vector(15 downto 0); -- data word from read command
  STATUS_OUT      : out std_logic_vector(7 downto 0); -- status and error bits
  VALID_OUT       : out std_logic;
  I2C_BUSY_OUT    : out std_logic;
  I2C_DONE_OUT    : out std_logic;
  -- I2C connections
  SDA_IN          : in  std_logic; 
  SDA_OUT         : out std_logic;
  SCL_IN          : in  std_logic;
  SCL_OUT         : out std_logic;
  -- Debug
  BSM_OUT         : out std_logic_vector(4 downto 0)
);
end i2c_slim;

architecture Behavioral of i2c_slim is

-- Signals
type STATES is (SLEEP, LOADA, GSTART, SENDA, LOADC, SENDC, LOADD, SENDD, GSTOP, INC,
				E_START, E_ADDR, E_CMD, E_WD, E_WD2, E_RSTART, E_RADDR, DONE, FAILED, CLRERR, LOADD2, SENDD2);
signal CURRENT_STATE, NEXT_STATE: STATES;

signal bsm          : std_logic_vector( 4 downto 0 );
signal phase        : std_logic; -- '0' => first phase, '1' => second phase of read cycle

signal start_x      : std_logic;
signal start        : std_logic; -- '0' => generate STOP, '1' => generate START
signal dostart_x    : std_logic;
signal dostart      : std_logic; -- trigger the GenStart module
signal dobyte_x     : std_logic;
signal dobyte       : std_logic; -- trigger the ByteSend module
signal i2c_done_x   : std_logic;
signal i2c_done     : std_logic; -- acknowledge signal to the outside world
signal running_x    : std_logic;
signal running      : std_logic; -- legacy
signal valid_x			: std_logic; 
signal valid				: std_logic;

signal load_dh_x    : std_logic;
signal load_dh      : std_logic;
signal load_dl_x    : std_logic;
signal load_dl      : std_logic;

signal sdone        : std_logic; -- acknowledge signal from GenStart module
signal sok          : std_logic; -- status signal from GenStart module
signal bdone        : std_logic; -- acknowledge signal from SendByte module
signal bok          : std_logic; -- status signal from SendByte module
signal e_sf         : std_logic; -- Start failed
signal e_anak       : std_logic; -- Adress byte NAK
signal e_cnak       : std_logic; -- Command byte NAK
signal e_dnak       : std_logic; -- First data byte NAK
signal e_dnak2      : std_logic; -- Second data byte NAK
signal e_rsf        : std_logic; -- Repeated Start failed
signal e_ranak      : std_logic; -- Repeated Adress NAK
signal i2c_byte     : std_logic_vector(8 downto 0);
signal i2c_dr       : std_logic_vector(8 downto 0);

signal i2c_drw			: std_logic_vector(15 downto 0);

signal s_scl        : std_logic;
signal r_scl        : std_logic;
signal s_sda        : std_logic;
signal r_sda        : std_logic;
signal r_scl_gs     : std_logic;
signal s_scl_gs     : std_logic;
signal r_sda_gs     : std_logic;
signal s_sda_gs     : std_logic;
signal r_scl_sb     : std_logic;
signal s_scl_sb     : std_logic;
signal r_sda_sb     : std_logic;
signal s_sda_sb     : std_logic;

signal scl_q				: std_logic_vector(2 downto 0);
signal sda_q				: std_logic_vector(2 downto 0);

signal i2c_speed    : std_logic_vector(8 downto 0);

-- Components
component i2c_gstart is
port( 
	CLK_IN				: in    std_logic;
	RESET_IN			: in    std_logic;
	START_IN			: in    std_logic;
	DOSTART_IN		: in    std_logic;
	I2C_SPEED_IN	: in    std_logic_vector(8 downto 0);
	SDONE_OUT			: out   std_logic;
	SOK_OUT				: out   std_logic;
	SDA_IN				: in    std_logic;
	SCL_IN				: in    std_logic;
	R_SCL_OUT			: out   std_logic;
	S_SCL_OUT			: out   std_logic;
	R_SDA_OUT			: out   std_logic;
	S_SDA_OUT			: out   std_logic;
	BSM_OUT				: out   std_logic_vector(3 downto 0)
);
end component i2c_gstart;

component i2c_sendb is
port( 
	CLK_IN				: in    std_logic;
	RESET_IN			: in    std_logic;
	DOBYTE_IN			: in    std_logic;
	I2C_SPEED_IN	: in    std_logic_vector(8 downto 0);
	I2C_BYTE_IN		: in    std_logic_vector(8 downto 0);
	I2C_BACK_OUT	: out   std_logic_vector(8 downto 0);
	SDA_IN				: in    std_logic;
	R_SDA_OUT			: out   std_logic;
	S_SDA_OUT			: out   std_logic;
--	SCL_IN			: in    std_logic;
	R_SCL_OUT			: out   std_logic;
	S_SCL_OUT			: out   std_logic;
	BDONE_OUT			: out   std_logic;
	BOK_OUT				: out   std_logic;
	BSM_OUT				: out   std_logic_vector(3 downto 0)
);
end component i2c_sendb;

begin

THE_SYNC_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		sda_q(2 downto 0) <= sda_q(1 downto 0) & SDA_IN;
		scl_q(2 downto 0) <= scl_q(1 downto 0) & SCL_IN;		
	end if;
end process THE_SYNC_PROC;

-- lower limit of speed
i2c_speed <= i2c_speed_in & "000";

-- Read phase indicator
THE_PHASE_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if( RESET = '1' ) then
			phase <= '0';
		elsif( CURRENT_STATE = INC ) then
			phase <= '1';
		elsif( (CURRENT_STATE = DONE) or (CURRENT_STATE = SLEEP) ) then
			phase <= '0';
		end if;
	end if;
end process THE_PHASE_PROC;

-- The main state machine
-- State memory process
STATE_MEM: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if( RESET = '1' ) then
			CURRENT_STATE <= SLEEP;
			start         <= '0';
			dostart       <= '0';
			dobyte        <= '0';
			i2c_done      <= '0';
			running       <= '0';
			load_dh       <= '0';
			load_dl       <= '0';
			valid         <= '0';
		else
			CURRENT_STATE <= NEXT_STATE;
			start         <= start_x;
			dostart       <= dostart_x;
			dobyte        <= dobyte_x;
			i2c_done      <= i2c_done_x;
			running       <= running_x;
			load_dh       <= load_dh_x;
			load_dl       <= load_dl_x;
			valid         <= valid_x;
		end if;
	end if;
end process STATE_MEM;

-- Transition matrix
TRANSFORM: process(CURRENT_STATE, i2c_go_in, sdone, sok, phase, bdone, bok, action_in, word_in)
begin
	NEXT_STATE <= SLEEP;
	start_x    <= '0';
	dostart_x  <= '0';
	dobyte_x   <= '0';
	i2c_done_x <= '0';
	running_x  <= '1';
	load_dh_x  <= '0';
	load_dl_x  <= '0';
	valid_x    <= '0';
	case CURRENT_STATE is
		when SLEEP      =>  
				if( i2c_go_in = '1' ) then
					NEXT_STATE <= CLRERR;
				else
					NEXT_STATE <= SLEEP;
					running_x  <= '0';
				end if;
		when CLRERR     =>  
				NEXT_STATE <= LOADA;
		when LOADA      =>  
				NEXT_STATE <= GSTART;
				start_x    <= '1';
				dostart_x  <= '1';
		when GSTART     =>  
				if   ( (sdone = '1') and (sok = '1') ) then
					NEXT_STATE <= SENDA; -- generating START did succeed
					dobyte_x   <= '1';
				elsif( (sdone = '1') and (sok = '0') and (phase = '0') ) then
					NEXT_STATE <= E_START; -- first START condition failed
				elsif( (sdone = '1') and (sok = '0') and (phase = '1') ) then
					NEXT_STATE <= E_RSTART; -- second START condition failed
				else
					NEXT_STATE <= GSTART; -- wait for START generation ending
					start_x    <= '1';
					dostart_x  <= '1';
				end if;
		when E_START    =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when E_RSTART   =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when SENDA      =>  
				if   ( (bdone = '1') and (bok = '1') and (action_in = '0') ) then
					NEXT_STATE <= LOADC; -- I2C write, send command
				elsif( (bdone = '1') and (bok = '1') and (action_in = '1') and (phase = '0') ) then
					NEXT_STATE <= LOADC; -- I2C read, send command
				elsif( (bdone = '1') and (bok = '1') and (action_in = '1') and (phase = '1') ) then
					NEXT_STATE <= LOADD; -- I2C read, send 0xff dummy byte
				elsif( (bdone = '1') and (bok = '0') and (phase = '0') ) then
					NEXT_STATE <= E_ADDR; -- first address phase failed
				elsif( (bdone = '1') and (bok = '0') and (phase = '1') ) then
					NEXT_STATE <= E_RADDR; -- second address phase failed
				else
					NEXT_STATE <= SENDA; -- wait for send address ending
					dobyte_x   <= '1';
				end if;
		when E_ADDR     =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when E_RADDR    =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when LOADC      =>  
				NEXT_STATE <= SENDC;
		when SENDC      =>  
				if   ( (bdone = '1') and (bok = '1') and (action_in = '0') and (word_in = '0') ) then
					NEXT_STATE <= LOADD2; -- I2C byte write, prepare data
				elsif( (bdone = '1') and (bok = '1') and (action_in = '0') and (word_in = '1') ) then
					NEXT_STATE <= LOADD; -- I2C word write, prepare data
				elsif( (bdone = '1') and (bok = '1') and (action_in = '1') ) then
					NEXT_STATE <= GSTOP; -- I2C read, first phase ends
					dostart_x  <= '1';
				elsif( (bdone = '1') and (bok = '0') ) then
					NEXT_STATE <= E_CMD; -- command phase failed
				else
					NEXT_STATE <= SENDC; -- wait for send command ending
					dobyte_x   <= '1';
				end if;
		when E_CMD      =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when LOADD      =>  
				NEXT_STATE <= SENDD;
		when SENDD      =>  
				if   ( (bdone = '1') and (bok = '1') and (action_in = '0') and (word_in = '0') ) then
					NEXT_STATE <= GSTOP;    -- I2C write, byte access, done
					dostart_x  <= '1';
				elsif( (bdone = '1') and (bok = '1') and (action_in = '0') and (word_in = '1') ) then
					NEXT_STATE <= LOADD2;    -- I2C write, word access, last byte to send
				elsif( (bdone = '1') and                 (action_in = '1') and (word_in = '0') ) then
					NEXT_STATE <= GSTOP; -- I2C read, byte access, data phase
					dostart_x  <= '1';
					load_dl_x  <= '1';
				elsif( (bdone = '1') and                 (action_in = '1') and (word_in = '1') ) then
					NEXT_STATE <= LOADD2; -- I2C read, word access, last byte to receive
					load_dh_x  <= '1';
				elsif( (bdone = '1') and (bok = '0') and (action_in = '0') ) then
					NEXT_STATE <= E_WD; -- I2C write, first data phase failed
				else
					NEXT_STATE <= SENDD; -- wait for send data ending
					dobyte_x   <= '1';
				end if;
		when LOADD2     =>  
				NEXT_STATE <= SENDD2;
		when SENDD2     =>  
				if   ( (bdone = '1') and (bok = '1') and (action_in = '0') ) then
					NEXT_STATE <= GSTOP; -- I2C write, done
					dostart_x  <= '1';
				elsif( (bdone = '1') and                 (action_in = '1') ) then
					NEXT_STATE <= GSTOP; -- I2C read, data phase
					dostart_x  <= '1';
					load_dl_x  <= '1';
				elsif( (bdone = '1') and (bok = '0') and (action_in = '0') ) then
					NEXT_STATE <= E_WD2; -- I2C write, data phase failed
				else
					NEXT_STATE <= SENDD2;
					dobyte_x   <= '1';
				end if;
		when E_WD2      =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when E_WD       =>  
				NEXT_STATE <= FAILED;
				dostart_x  <= '1';
		when GSTOP      =>  
				if   ( (sdone = '1') and (action_in = '0') ) then
					NEXT_STATE <= DONE;
					i2c_done_x <= '1';
					valid_x    <= '1';
				elsif( (sdone = '1') and (action_in = '1') and (phase = '1') ) then
					NEXT_STATE <= DONE;
					i2c_done_x <= '1';
					valid_x    <= '1';
				elsif( (sdone = '1') and (action_in = '1') and (phase = '0') ) then
					NEXT_STATE <= INC;
				else
					NEXT_STATE <= GSTOP;
					dostart_x  <= '1';
				end if;
		when INC        =>  
				NEXT_STATE <= LOADA;
		when FAILED     =>  
				if( sdone = '1' ) then
						NEXT_STATE <= DONE;
						i2c_done_x <= '1';
						running_x  <= '0';
--						valid_x    <= '1';
					else
						NEXT_STATE <= FAILED;
						dostart_x  <= '1';
					end if;
		when DONE       =>  
				if( i2c_go_in = '1' ) then
						NEXT_STATE <= DONE;
						i2c_done_x <= '1';
						running_x  <= '0';
					else
						NEXT_STATE <= SLEEP;
						running_x  <= '0';
					end if;
				-- Just in case...
		when others     =>  
				NEXT_STATE <= SLEEP;
	end case;
end process TRANSFORM;

-- Output decoding
DECODE: process(CURRENT_STATE)
begin
	case CURRENT_STATE is
		when SLEEP      =>  bsm <= b"00000"; -- 00
		when LOADA      =>  bsm <= b"00001"; -- 01
		when GSTART     =>  bsm <= b"00010"; -- 02
		when SENDA      =>  bsm <= b"00011"; -- 03
		when LOADC      =>  bsm <= b"00100"; -- 04
		when SENDC      =>  bsm <= b"00101"; -- 05
		when LOADD      =>  bsm <= b"00110"; -- 06
		when SENDD      =>  bsm <= b"00111"; -- 07
		when GSTOP      =>  bsm <= b"01000"; -- 08
		when INC        =>  bsm <= b"01001"; -- 09
		when FAILED     =>  bsm <= b"01010"; -- 0a
		when DONE       =>  bsm <= b"01011"; -- 0b
		when CLRERR     =>  bsm <= b"01100"; -- 0c
		when LOADD2			=>	bsm <= b"01101"; -- 0d
		when SENDD2			=>	bsm <= b"01110"; -- 0e
		when E_START    =>  bsm <= b"10000"; -- 10
		when E_RSTART   =>  bsm <= b"10001"; -- 11
		when E_ADDR     =>  bsm <= b"10010"; -- 12
		when E_RADDR    =>  bsm <= b"10011"; -- 13
		when E_CMD      =>  bsm <= b"10100"; -- 14
		when E_WD       =>  bsm <= b"10101"; -- 15
		when E_WD2      =>  bsm <= b"10110"; -- 16
		when others     =>  bsm <= b"11111"; -- 1f
	end case;
end process DECODE;

-- We need to load different data sets
--LOAD_DATA_PROC: process( CLOCK, RESET, CURRENT_STATE, action_in, phase)
LOAD_DATA_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if   ( RESET = '1' ) then
			i2c_byte <= (others => '1');
		elsif( (CURRENT_STATE = LOADA) and (phase = '0') ) then
			i2c_byte <= i2c_addr_in(7 downto 1) & '0' & '1'; -- send write address, receive ACK
		elsif( (CURRENT_STATE = LOADA) and (phase = '1') ) then
			i2c_byte <= i2c_addr_in(7 downto 1) & '1' & '1'; -- send read address, receive ACK
		elsif( (CURRENT_STATE = LOADC) and (action_in = '0') ) then
			i2c_byte <= i2c_cmd_in(7 downto 0) & '1'; -- send command byte (WRITE), receive ACK
		elsif( (CURRENT_STATE = LOADC) and (action_in = '1') ) then
			i2c_byte <= i2c_cmd_in(7 downto 0) & '1'; -- send command byte (READ), receive ACK
		elsif( (CURRENT_STATE = LOADD) and (action_in = '0') ) then
			i2c_byte <= i2c_dw_in(15 downto 8) & '1'; -- send data byte, receive ACK
		elsif( (CURRENT_STATE = LOADD2) and (action_in = '0') ) then
			i2c_byte <= i2c_dw_in(7 downto 0) & '1'; -- send data byte, receive ACK
		elsif( (CURRENT_STATE = LOADD) and (action_in = '1') ) then
			i2c_byte <= x"ff" & '0'; -- send 0xff byte, send ACK
		elsif( (CURRENT_STATE = LOADD2) and (action_in = '1') ) then
			i2c_byte <= x"ff" & '1'; -- send 0xff byte, send NACK
		end if;
	end if;
end process LOAD_DATA_PROC;

-- The SendByte module
THE_I2C_SENDB: I2C_SENDB
port map(
	CLK_IN          => CLOCK,
	RESET_IN        => RESET,
	DOBYTE_IN       => dobyte,
	I2C_SPEED_IN    => i2c_speed,
	I2C_BYTE_IN     => i2c_byte,
	I2C_BACK_OUT    => i2c_dr,
	SDA_IN          => sda_q(2), -- changed
	R_SDA_OUT       => r_sda_sb,
	S_SDA_OUT       => s_sda_sb,
--  SCL_IN          => scl_q(2), -- changes
	R_SCL_OUT       => r_scl_sb,
	S_SCL_OUT       => s_scl_sb,
	BDONE_OUT       => bdone,
	BOK_OUT         => bok,
	BSM_OUT         => open
);

-- The GenStart module
THE_I2C_GSTART: I2C_GSTART
port map(
	CLK_IN          => CLOCK,
	RESET_IN        => RESET,
	START_IN        => start,
	DOSTART_IN      => dostart,
	I2C_SPEED_IN    => i2c_speed,
	SDONE_OUT       => sdone,
	SOK_OUT         => sok,
	SDA_IN          => sda_q(2), -- changed
	SCL_IN          => scl_q(2), -- changed
	R_SCL_OUT       => r_scl_gs,
	S_SCL_OUT       => s_scl_gs,
	R_SDA_OUT       => r_sda_gs,
	S_SDA_OUT       => s_sda_gs,
	BSM_OUT         => open
);

r_scl <= r_scl_gs or r_scl_sb;
s_scl <= s_scl_gs or s_scl_sb;
r_sda <= r_sda_gs or r_sda_sb;
s_sda <= s_sda_gs or s_sda_sb;

-- Output flipflop for SCL line
THE_SCL_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if( RESET = '1' ) then
			SCL_OUT <= '1';
		elsif( (r_scl = '1') and (s_scl = '0') ) then
			SCL_OUT <= '0';
		elsif( (r_scl = '0') and (s_scl = '1') ) then
			SCL_OUT <= '1';
		end if;
	end if;
end process THE_SCL_PROC;

-- Output flipflop for SDA line
THE_SDA_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if( RESET = '1' ) then
			SDA_OUT <= '1';
		elsif( (r_sda = '1') and (s_sda = '0') ) then
			SDA_OUT <= '0';
		elsif( (r_sda = '0') and (s_sda = '1') ) then
			SDA_OUT <= '1';
		end if;
	end if;
end process THE_SDA_PROC;

-- Error bits
THE_ERR_REG_PROC: process( CLOCK )
begin
	if( rising_edge(CLOCK) ) then
		if( RESET = '1' ) then
			e_sf        <= '0';
			e_anak      <= '0';
			e_cnak      <= '0';
			e_dnak      <= '0';
			e_dnak2     <= '0';
			e_rsf       <= '0';
			e_ranak     <= '0';
		elsif( CURRENT_STATE = CLRERR ) then
			e_sf        <= '0';
			e_anak      <= '0';
			e_cnak      <= '0';
			e_dnak      <= '0';
			e_dnak2     <= '0';
			e_rsf       <= '0';
			e_ranak     <= '0';
		elsif( CURRENT_STATE = E_START ) then
			e_sf        <= '1';
		elsif( CURRENT_STATE = E_RSTART ) then
			e_rsf       <= '1';
		elsif( CURRENT_STATE = E_ADDR ) then
			e_anak      <= '1';
		elsif( CURRENT_STATE = E_RADDR ) then
			e_ranak     <= '1';
		elsif( CURRENT_STATE = E_CMD ) then
			e_cnak      <= '1';
		elsif( CURRENT_STATE = E_WD ) then
			e_dnak      <= '1';
		elsif( CURRENT_STATE = E_WD2 ) then
			e_dnak2     <= '1';
		end if;
	end if;
end process THE_ERR_REG_PROC;

-- store data read from I2C (high byte)
THE_STORE_READ_H_PROC: process( CLOCK ) 
begin
	if( rising_edge(CLOCK) ) then
		if   ( RESET = '1' ) then
			i2c_drw(15 downto 8) <= (others => '0');
		elsif( rising_edge(CLOCK) ) then
			if( load_dh = '1' ) then
				i2c_drw(15 downto 8) <= i2c_dr(8 downto 1);
			end if;
		end if;
	end if;
end process THE_STORE_READ_H_PROC;

-- store data read from I2C (low byte)
THE_STORE_READ_L_PROC: process( CLOCK ) 
begin
	if( rising_edge(CLOCK) ) then
		if   ( RESET = '1' ) then
			i2c_drw(7 downto 0) <= (others => '0');
		elsif( rising_edge(CLOCK) ) then
			if( load_dl = '1' ) then
				i2c_drw(7 downto 0) <= i2c_dr(8 downto 1);
			end if;
		end if;
	end if;
end process THE_STORE_READ_L_PROC;

-- Status output
status_out(7) <= e_sf;
status_out(6) <= e_rsf;
status_out(5) <= e_anak;
status_out(4) <= e_ranak;
status_out(3) <= e_cnak;
status_out(2) <= e_dnak;
status_out(1) <= e_dnak2;
status_out(0) <= '0';

-- Outputs
I2C_DR_OUT      <= i2c_drw;
I2C_BUSY_OUT    <= running;
I2C_DONE_OUT    <= i2c_done;
VALID_OUT       <= valid;
BSM_OUT         <= bsm;

end Behavioral;
