-- Generic flash controller
-- Should be placed between spi_slave and the local logic
-- (see e.g. dirich/thresholds/thresholds.vhd)
--
-- Data in the user flash will be unpacked during boot, the data has the
-- following format:
--   Version : 1 Byte (must be 0x01 or 0x02, otherwise the unpacking will be cancelled)
--   Version = 0x01 : 16 Bit version
--     SPI address : 1 Byte
--     SPI data : 2 Bytes in big-endian
--   Version = 0x02 : 32 Bit version
--     SPI address : 1 Byte
--     SPI data : 4 Bytes in big-endian
--     2 paddings bytes
--
-- SPI registers:
--  0x4x : Mapped flash page (16 Bytes)
--  0x50 : Flash command
--         Mode=0: (see Padiwa manual)
--           Bit 15-13 : Flash command
--           Bit 12-0  : Flash page
--         Mode=1:
--           16 bit page adress
--  0x51 : Flash command (Mode 1)
--         Bit 3-1 : Contains the 3-bit flash command when in mode 1
--         The command must be written before the page address and can be re-used
--  0x5C : control register
--         Bit 0 : Enable cfg flash
--         Bit 1 : Flash error (read-only)
--         Bit 2 : Flash busy (read-only)
--         Bit 4 : Master start (starts unpacking = booting)
--         Bit 5 : Master running (read-only)
--         Bit 8 : Mode
--                 0: Old padiwa 13 bit mode
--                 1: New 16 bit mode
--  0x5D : Bits 0/1: Flash memory read buswidth:
--         00 : 8 Bit, 01 : 16 Bit, 11 : 32 Bit (if available)
--         Bit 4: 0: Little Endian; 1: Big Endian
--         Bit 11-8 : Burst counter
--  0x5E-5F : Debug registers


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library machxo3lf;
use machxo3lf.all;

library work;
use work.trb_net_std.all;

entity generic_flash_ctrl is
  generic (MASTER_STARTPAGE : std_logic_vector(12 downto 0) := "1" & x"C00";
           USE_OLD_13BIT_MODE : integer := c_NO;
           DATA_BUS_WIDTH   : integer := 16;
           CFG_FLASH_DISABLE_COUNTER_WIDTH   : integer := 32; --42sec@100MHz
           USE_I2C_PROG : integer := c_YES
           );
  port(
    CLK_l : in  std_logic; --local clock
    CLK_f : in  std_logic; --flash clock
    RESET : in  std_logic;

    -- SPI/UART in host direction
    SPI_DATA_IN   : in  std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
    SPI_DATA_OUT  : out std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
    SPI_ADDR_IN   : in  std_logic_vector(7 downto 0);
    SPI_WRITE_IN  : in  std_logic;
    SPI_READ_IN   : in  std_logic;
    SPI_READY_OUT : out std_logic;
    SPI_BUSY_IN   : in  std_logic;
        
    -- SPI/UART in local direction
    LOC_DATA_OUT  : out std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
    LOC_DATA_IN   : in  std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
    LOC_ADDR_OUT  : out std_logic_vector(7 downto 0);
    LOC_WRITE_OUT : out std_logic;
    LOC_READ_OUT  : out std_logic;
    LOC_READY_IN  : in  std_logic;
    LOC_BUSY_OUT  : out std_logic;
    
    SCL           : inout std_logic;
    SDA           : inout std_logic
    );
end entity;

architecture arch of generic_flash_ctrl is

  component UFM_WB
    port(
      clk_i : in std_logic;
      rst_n : in std_logic;
      cmd       : in std_logic_vector(2 downto 0);
      ufm_page  : in std_logic_vector(15 downto 0);
      select_cfg : in std_logic;
      GO        : in std_logic;
      BUSY      : out std_logic;
      ERR       : out std_logic;
      mem_clk   : out std_logic;
      mem_we    : out std_logic;
      mem_ce    : out std_logic;
      mem_addr  : out std_logic_vector(3 downto 0);
      mem_wr_data : out std_logic_vector(7 downto 0);
      mem_rd_data : in  std_logic_vector(7 downto 0)
      );
  end component;
  
  component UFM_WB_I2C_Prog
    port(
      clk_i : in std_logic;
      rst_n : in std_logic;
      cmd       : in std_logic_vector(2 downto 0);
      ufm_page  : in std_logic_vector(15 downto 0);
      select_cfg : in std_logic;
      GO        : in std_logic;
      BUSY      : out std_logic;
      ERR       : out std_logic;
      mem_clk   : out std_logic;
      mem_we    : out std_logic;
      mem_ce    : out std_logic;
      mem_addr  : out std_logic_vector(3 downto 0);
      mem_wr_data : out std_logic_vector(7 downto 0);
      mem_rd_data : in  std_logic_vector(7 downto 0);
      i2c_scl     : inout std_logic;
      i2c_sda     : inout std_logic
      );
  end component;

  signal reg_SPI_DATA_OUT  : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal reg_SPI_READY_OUT : std_logic;
  signal reg_LOC_DATA_OUT  : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal reg_LOC_ADDR_OUT  : std_logic_vector(7 downto 0);
  signal reg_LOC_WRITE_OUT : std_logic;
  signal reg_LOC_READ_OUT  : std_logic;

  signal reg_SPI_DATA_IN : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal reg_SPI_ADDR_IN : std_logic_vector(7 downto 0);
  signal reg_SPI_READ_IN : std_logic;
  signal reg_SPI_WRITE_IN : std_logic;
  
  signal flashram_addr_i : std_logic_vector(3 downto 0);
  signal flashram_cen_i  : std_logic;
  signal flashram_reset  : std_logic;
  signal flashram_write_i: std_logic;
  signal flashram_data_i : std_logic_vector(7 downto 0);
  signal flashram_data_o : std_logic_vector(7 downto 0);

  signal flash_command : std_logic_vector(2 downto 0);
  signal flash_command_register : std_logic_vector(2 downto 0) := "000";
  signal spi_flash_command : std_logic_vector(2 downto 0);
  signal flash_page    : std_logic_vector(15 downto 0);
  signal spi_flash_page    : std_logic_vector(15 downto 0);
  signal flash_go      : std_logic;
  signal spi_flash_go  : std_logic;
  signal flash_busy    : std_logic;
  signal flash_err     : std_logic;
  signal select_cfg : std_logic;
  signal flash_mode : std_logic := '1';
  
  signal ram_write_i : std_logic;
  signal spi_ram_write_i : std_logic;
  signal master_ram_write_i : std_logic;
  signal ram_data_i  : std_logic_vector(7 downto 0);
  signal spi_ram_data_i  : std_logic_vector(7 downto 0);
  signal master_ram_data_i  : std_logic_vector(7 downto 0);
  signal ram_data_o  : std_logic_vector(7 downto 0);
  signal raw_ram_data_o  : std_logic_vector(7 downto 0);
  signal ram_addr_i  : std_logic_vector(3 downto 0);
  signal spi_ram_addr_i  : std_logic_vector(3 downto 0);
  
  signal enable_cfg_flash : std_logic_vector(2 downto 0);
  signal cfg_flash_disable_counter : std_logic_vector(CFG_FLASH_DISABLE_COUNTER_WIDTH downto 0);
  signal testreg : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal memreg  : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal burst_counter : std_logic_vector(3 downto 0);

  signal out_delay : std_logic_vector(2 downto 0);
  

  type state_type is (IDLE, EnableFLASH1, EnableFLASH2, EnableFLASH3, Start, ReadPage, WritePage, WaitFlash1, WaitFlash2,
                      EraseFLASH1, EraseFLASH2, EraseFLASH3,
                      ReadSPI0, ReadSPI1, ReadSPI2, ReadSPI3,
                      ReadRAM, ReadRAM0, WaitRAM1, WaitRAM2, WaitRAM3, WaitRAM4,
                      WaitRAM2_32, WaitRAM3_32, WaitRAM4_32, WaitRAM5_32, WaitRAM6_32, WriteSPI, DisableFLASH1, DisableFLASH2, DisableFLASH3);
  signal state : state_type := IDLE;
  signal master_running       : std_logic;
  signal master_word_counter  : std_logic_vector(3 downto 0);
  signal master_flash_page    : std_logic_vector(12 downto 0) := "0000000000000";
  signal master_flash_command : std_logic_vector(2 downto 0);
  signal master_flash_go      : std_logic;
  signal master_start_reg     : std_logic := '0';
  signal clean_master_start_reg : std_logic := '0';
  signal master_readback      : std_logic := '0';
  signal master_erase         : std_logic := '0';
  signal master_DATA_OUT  : std_logic_vector(DATA_BUS_WIDTH-1 downto 0);
  signal master_ADDR_OUT  : std_logic_vector(7 downto 0);
  signal master_WRITE_OUT : std_logic;
  signal master_READ_OUT  : std_logic;
  signal master_readback_reg : std_logic_vector(7 downto 0) := x"00";
  signal readback_reg : std_logic_vector(7 downto 0) := x"00";
  signal master_readback_num : std_logic_vector(4 downto 0) := "00000";
  signal readback_num : std_logic_vector(3 downto 0) := x"0";
  
  
  
  signal auto_reset : std_logic := '1';
  signal auto_dbg : std_logic := '0';
  signal auto_cnt : std_logic_vector(19 downto 0) := x"00000";
  
begin
  
  THE_FLASH_RAM : entity work.flashram
    port map(
      DataInA   => ram_data_i,
      DataInB   => flashram_data_i,
      AddressA  => ram_addr_i,
      AddressB  => flashram_addr_i,
      ClockA    => CLK_l,
      ClockB    => CLK_f,
      ClockEnA  => '1',
      ClockEnB  => flashram_cen_i,
      WrA       => ram_write_i,
      WrB       => flashram_write_i,
      ResetA    => RESET,
      ResetB    => RESET,
      QA        => raw_ram_data_o,
      QB        => flashram_data_o
      );

  GEN_UFM_WB : if USE_I2C_PROG = 0 generate
    THE_FLASH : UFM_WB
      port map(
        clk_i      => CLK_f,
        rst_n      => not RESET,
        cmd        => flash_command,
        ufm_page   => flash_page,
        select_cfg => select_cfg,
        GO         => flash_go,
        BUSY       => flash_busy,
        ERR        => flash_err,
        mem_clk    => open,
        mem_we      => flashram_write_i,
        mem_ce      => flashram_cen_i,
        mem_addr    => flashram_addr_i,
        mem_wr_data => flashram_data_i,
        mem_rd_data => flashram_data_o
      );
      
      SCL <= 'Z';
      SDA <= 'Z';
   end generate GEN_UFM_WB;
      
  GEN_UFM_WB_I2C : if USE_I2C_PROG = 1 generate
    THE_FLASH : UFM_WB_I2C_Prog
      port map(
        clk_i      => CLK_f,
        rst_n      => not RESET,
        cmd        => flash_command,
        ufm_page   => flash_page,
        select_cfg => select_cfg,
        GO         => flash_go,
        BUSY       => flash_busy,
        ERR        => flash_err,
        mem_clk    => open,
        mem_we      => flashram_write_i,
        mem_ce      => flashram_cen_i,
        mem_addr    => flashram_addr_i,
        mem_wr_data => flashram_data_i,
        mem_rd_data => flashram_data_o,
        i2c_scl     => SCL,
        i2c_sda     => SDA
      );
  end generate GEN_UFM_WB_I2C;
  
  LOC_DATA_OUT  <= reg_LOC_DATA_OUT;
  LOC_ADDR_OUT  <= reg_LOC_ADDR_OUT;
  LOC_WRITE_OUT <= reg_LOC_WRITE_OUT;
  LOC_READ_OUT  <= reg_LOC_READ_OUT;
  SPI_DATA_OUT  <= reg_SPI_DATA_OUT;
  SPI_READY_OUT <= reg_SPI_READY_OUT;

  LOC_BUSY_OUT  <= SPI_BUSY_IN;

MODE_SELECTOR : process 
begin
  wait until rising_edge(clk_f);
  if (master_running = '1') then
    flash_go      <= master_flash_go;
    flash_command <= master_flash_command;
  else
    flash_go      <= spi_flash_go;
    flash_command <= spi_flash_command;
  end if;

  if (flash_mode = '0' and USE_OLD_13BIT_MODE = c_YES) then
    select_cfg <= '0';
    if (master_running = '1') then
      flash_page(9 downto 0)   <= master_flash_page(9 downto 0);
      flash_page(15 downto 10) <= "000000";
    else
      flash_page(12 downto 0)  <= spi_flash_page(12 downto 0);
      flash_page(15 downto 13) <= "000";
      select_cfg <= '1';
      if (spi_flash_page(12 downto 10) = "111" or enable_cfg_flash(0) = '0') then
        select_cfg <= '0';
        flash_page(12 downto 10) <= "000";
      end if;
    end if;
  else --new 16 bit mode
    if (master_running = '1') then
      flash_page(9 downto 0)   <= master_flash_page(9 downto 0);
      flash_page(15 downto 10) <= "000000";
      select_cfg <= '0';
    else
      flash_page <= spi_flash_page;
      select_cfg <= enable_cfg_flash(0);
    end if;
  end if;
end process;                          

SYNC_RAM : process
begin
  wait until rising_edge(clk_l);
  ram_data_o <= raw_ram_data_o;
end process;

--reg_SPI : process begin
--  wait until rising_edge(clk_l);

--  reg_SPI_DATA_IN <= SPI_DATA_IN;
--  reg_SPI_ADDR_IN <= SPI_ADDR_IN;
--  reg_SPI_WRITE_IN <= SPI_WRITE_IN;
--  reg_SPI_READ_IN <= SPI_READ_IN;
--end process;
  
PROC_SELECTOR : process begin
  wait until rising_edge(clk_l);

  reg_LOC_DATA_OUT  <= SPI_DATA_IN;
  reg_LOC_ADDR_OUT  <= SPI_ADDR_IN;
  reg_LOC_WRITE_OUT <= SPI_WRITE_IN;
  reg_LOC_READ_OUT  <= SPI_READ_IN;
  if (out_delay = "000") then
    reg_SPI_DATA_OUT  <= LOC_DATA_IN;
    reg_SPI_READY_OUT <= LOC_READY_IN;
  end if;
  
  spi_ram_write_i <= '0';
  spi_ram_data_i <= x"00";
  if (burst_counter = "0000" and out_delay = "000") then
    spi_ram_addr_i <= x"0";
  else
    spi_ram_addr_i <= spi_ram_addr_i;
  end if;

  if (flash_busy = '1') then
    spi_flash_go <= '0';
  end if;
      
  if (clean_master_start_reg = '1') then
    master_start_reg <= '0';
  end if;

  if enable_cfg_flash(2 downto 0) = "000" then
    cfg_flash_disable_counter <= (others => '0');
  elsif enable_cfg_flash(2 downto 0) = "001" and cfg_flash_disable_counter(CFG_FLASH_DISABLE_COUNTER_WIDTH) = '1' then
    --timeout
    enable_cfg_flash(2 downto 0) <= "000";
  elsif enable_cfg_flash(2 downto 0) = "001" then
    cfg_flash_disable_counter <= std_logic_vector(unsigned(cfg_flash_disable_counter)+1);
  end if;

  if master_start_reg = '1' and master_readback = '1' then
    enable_cfg_flash(2 downto 0) <= "000";
  end if;
  
  if (SPI_WRITE_IN = '1') then
    if (SPI_ADDR_IN(7 downto 4) = x"4") then
      reg_LOC_WRITE_OUT <= '0';
      spi_ram_write_i <= '1';
      spi_ram_data_i  <= SPI_DATA_IN(7 downto 0);
      spi_ram_addr_i  <= SPI_ADDR_IN(3 downto 0);
      cfg_flash_disable_counter <= (others => '0');
    elsif (SPI_ADDR_IN(7 downto 0) = x"5B") then
      master_start_reg <= SPI_DATA_IN(0);
      master_readback  <= SPI_DATA_IN(1);
      master_erase     <= SPI_DATA_IN(2);
      readback_num     <= SPI_DATA_IN(7 downto 4);
      readback_reg     <= SPI_DATA_IN(15 downto 8);
    elsif (SPI_ADDR_IN(7 downto 0) = x"5C") then
      reg_LOC_WRITE_OUT <= '0';
      if SPI_DATA_IN(0) = '0' then
        enable_cfg_flash(2 downto 0) <= "000";
      elsif SPI_DATA_IN(2 downto 0) = "011" and (enable_cfg_flash(2 downto 0) = "000" or enable_cfg_flash(2 downto 0) = "001") then --stage1
        enable_cfg_flash(2 downto 0) <= "010";
      elsif SPI_DATA_IN(2 downto 0) = "101" and enable_cfg_flash(2 downto 0) = "010" then --stage2
        enable_cfg_flash(2 downto 0) <= "100";
      elsif SPI_DATA_IN(2 downto 0) = "001" and enable_cfg_flash(2 downto 0) = "100" then --stage3
        enable_cfg_flash(2 downto 0) <= "001";
      else
        enable_cfg_flash(2 downto 0) <= "000";
      end if;        
      --flash_mode       <= SPI_DATA_IN(8);  --XXX
    elsif (SPI_ADDR_IN(7 downto 0) = x"5d") then
      reg_LOC_WRITE_OUT <= '0';
      memreg <= SPI_DATA_IN;
    elsif (SPI_ADDR_IN(7 downto 0) = x"5f") then
      reg_LOC_WRITE_OUT <= '0';
      testreg <= SPI_DATA_IN;
    elsif (SPI_ADDR_IN(7 downto 0) = x"51") then
      flash_command_register <= SPI_DATA_IN(3 downto 1);
    elsif (SPI_ADDR_IN(7 downto 0) = x"50" and spi_flash_go = '0') then
      reg_LOC_WRITE_OUT <= '0';
      if (flash_mode = '0' and USE_OLD_13BIT_MODE = c_YES) then
        spi_flash_command <= SPI_DATA_IN(15 downto 13);
        if (enable_cfg_flash(0) = '1') then
          spi_flash_page <= "000" & SPI_DATA_IN(12 downto 0);
        else
          spi_flash_page <= "000111" & SPI_DATA_IN(9 downto 0);
        end if;
      else
        spi_flash_command <= flash_command_register;
        spi_flash_page    <= SPI_DATA_IN(15 downto 0);
      end if;
      spi_flash_go <= '1';
    end if;
  end if;

  if (master_running = '1') then
    reg_LOC_WRITE_OUT <= master_WRITE_OUT;
    reg_LOC_READ_OUT  <= master_READ_OUT;
    reg_LOC_DATA_OUT  <= master_DATA_OUT;
    reg_LOC_ADDR_OUT  <= master_ADDR_OUT;
  end if;

  if (out_delay = "001") then
    reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT <= '0';
    if (SPI_BUSY_IN = '0') then
      if (memreg(0) = '1') then
        -- at least 16 bit burst
        spi_ram_addr_i <= std_logic_vector(unsigned(spi_ram_addr_i)+1); -- prepare nibble2
      end if;      
      out_delay <= "010";
    end if;
  elsif (out_delay = "010") then
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT  <= '0';
    out_delay <= "011";
    if (memreg(1) = '1') then
      -- at least 24 bit burst
      spi_ram_addr_i <= std_logic_vector(unsigned(spi_ram_addr_i)+1); --prepare nibble3
    end if;
  elsif (out_delay = "011") then
    reg_LOC_READ_OUT  <= '0';
    if (memreg(0) = '0') then
      out_delay <= "111";
      reg_SPI_DATA_OUT(15 downto 0) <= flash_busy & flash_err & "000000" & ram_data_o; -- write nibble1
      reg_SPI_READY_OUT <= '1';
    else
      reg_SPI_READY_OUT <= '0';
      if (memreg(4) = '1') then
        reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto DATA_BUS_WIDTH-8) <= ram_data_o;
      else
        reg_SPI_DATA_OUT(15 downto 0) <= x"00" & ram_data_o; -- write nibble1
      end if;
      spi_ram_addr_i <= std_logic_vector(unsigned(spi_ram_addr_i)+1); --prepare nibble4
      out_delay <= "100";
    end if;
  elsif (out_delay = "100") then
    if (memreg(4) = '1') then -- write nibble2
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto DATA_BUS_WIDTH-16) <= reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto DATA_BUS_WIDTH-8) & ram_data_o;
    else
      reg_SPI_DATA_OUT(15 downto 0) <= ram_data_o & reg_SPI_DATA_OUT(7 downto 0);
    end if;
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT <= '0';
    out_delay <= "101";
    if (memreg(1) = '0' or DATA_BUS_WIDTH < 24) then
      out_delay <= "111";
      reg_SPI_READY_OUT <= '1';
    end if;
  elsif (out_delay = "101") then
    if (DATA_BUS_WIDTH > 16) then
      if (memreg(4) = '1') then -- write nibble3
        reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto DATA_BUS_WIDTH-24) <= reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto DATA_BUS_WIDTH-16) & ram_data_o;
      else
        reg_SPI_DATA_OUT(23 downto 0) <= ram_data_o & reg_SPI_DATA_OUT(15 downto 0);
      end if;
    end if;
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT <= '0';
    out_delay <= "110";    
    if (DATA_BUS_WIDTH < 32) then
      out_delay <= "111";
      reg_SPI_READY_OUT <= '1';
    end if;
  elsif (out_delay = "110") then
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT <= '1';
    if (DATA_BUS_WIDTH > 24) then
      if (memreg(4) = '1') then -- write nibble4
        reg_SPI_DATA_OUT(31 downto 0) <= reg_SPI_DATA_OUT(31 downto 8) & ram_data_o;
      else
        reg_SPI_DATA_OUT(31 downto 0) <= ram_data_o & reg_SPI_DATA_OUT(23 downto 0);
      end if;
    end if;
    out_delay <= "111";
  else
    out_delay <= "000";
  end if;

  if (out_delay = "000" and burst_counter /= "0000" and SPI_BUSY_IN = '0' and reg_SPI_READY_OUT = '0') then
    out_delay <= "001";
    reg_LOC_READ_OUT  <= '0';
    reg_SPI_READY_OUT <= '0';
    spi_ram_addr_i <= std_logic_vector(unsigned(spi_ram_addr_i)+1);
    burst_counter <= std_logic_vector(unsigned(burst_counter)-1);
  end if;
  
  if (SPI_READ_IN = '1') then    
    if (SPI_ADDR_IN(7 downto 4) = x"4" and burst_counter = "0000" and out_delay = "000") then
      cfg_flash_disable_counter <= (others => '0');
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      out_delay <= "001";
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '0';
      spi_ram_addr_i <= SPI_ADDR_IN(3 downto 0); -- prepare nibble1
      burst_counter <= memreg(11 downto 8);
    elsif (SPI_ADDR_IN(7 downto 0) = x"51") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      reg_SPI_DATA_OUT(2 downto 0) <= flash_command_register;
    elsif (SPI_ADDR_IN(7 downto 0) = x"5B") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      reg_SPI_DATA_OUT(15 downto 0) <= readback_reg & readback_num & master_running & master_erase & master_readback & master_start_reg;
    elsif (SPI_ADDR_IN(7 downto 0) = x"5C") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      reg_SPI_DATA_OUT(15 downto 0) <= x"0" & "000" & flash_mode & '0' & auto_reset & "00" & '0' & flash_busy & flash_err & enable_cfg_flash(0);
    elsif (SPI_ADDR_IN(7 downto 0) = x"5d") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT  <= memreg;
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      reg_SPI_DATA_OUT(11 downto 8) <= burst_counter;
      -- reg_SPI_DATA_OUT(15 downto 0) <= auto_dbg & "00" & master_flash_page;
    elsif (SPI_ADDR_IN(7 downto 0) = x"5e") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT(DATA_BUS_WIDTH-1 downto 0) <= (others => '0');
      reg_SPI_DATA_OUT  <= master_DATA_OUT;
    elsif (SPI_ADDR_IN(7 downto 0) = x"5f") then
      reg_LOC_READ_OUT  <= '0';
      reg_SPI_READY_OUT <= '1';
      reg_SPI_DATA_OUT  <= testreg;
      --reg_SPI_DATA_OUT(15 downto 0)  <= x"000" & master_word_counter;
    end if;
  end if;

end process;

ram_addr_i  <= master_word_counter when master_running = '1' else spi_ram_addr_i;
ram_data_i  <= master_ram_data_i   when master_running = '1' else spi_ram_data_i;
ram_write_i <= master_ram_write_i  when master_running = '1' else spi_ram_write_i;


MASTER_DRIVER :  process (clk_l)
begin
  if rising_edge(clk_l) then

    if (master_flash_go = '1' and flash_busy = '1') then
      master_flash_go  <= '0';
    end if;
      
    master_WRITE_OUT <= '0';
    master_READ_OUT  <= '0';
    clean_master_start_reg <= '0';
    master_running <= '1';
    
--    if (DATA_BUS_WIDTH-1 > 15) then
--      master_DATA_OUT(DATA_BUS_WIDTH-1 downto 16) <= (others => '0');
--    end if;
            
    case state is
      when IDLE =>
        master_running <= '0';
        --if ((master_start_reg = '1' or auto_reset = '1') and (flash_mode = '0')) then
        if ((master_start_reg = '1' or auto_reset = '1') and (flash_mode = '1')) then
          --state <= Start;
          --state <= EnableFLASH1;
          clean_master_start_reg <= '1';
          master_DATA_OUT <= (others => '0');
          master_ADDR_OUT <= x"00";
          if (auto_reset = '1') then
            auto_cnt <= std_logic_vector(unsigned(auto_cnt) + 1);
            state <= IDLE;
            if (auto_cnt = x"FFFFF") then
              auto_reset <= '0';
              state <= EnableFLASH1;
              master_flash_command <= "100";
              master_flash_go <= '1';
            end if;
          else
            --master_flash_command <= "000";
            state <= EnableFLASH1;
            master_flash_command <= "100";
            master_flash_go <= '1';
          end if;
        end if;
        
      when EnableFLASH1 =>
        auto_dbg <= '1';
        state <= EnableFLASH2;
      when EnableFLASH2 =>
        state <= EnableFLASH3;
      when EnableFLASH3 =>
        state <= EnableFLASH3;
        if (flash_busy = '0' and master_flash_go = '0') then
           state <= Start;
           master_flash_command <= "000";
           if master_erase = '1' then
             state <= EraseFlash1;
             master_flash_command <= "111";
             master_flash_go <= '1';
           end if;
        end if;

      when EraseFlash1 =>
        state <= EraseFlash2;
      when EraseFlash2 =>
        state <= EraseFlash3;
      when EraseFlash3 =>
        state <= EraseFlash3;
        if (flash_busy = '0' and master_flash_go = '0') then
          state <= Start;
          master_flash_command <= "000";
        end if;
        
      when Start =>
        --master_flash_page   <= MASTER_STARTPAGE;
        master_flash_page   <= (others => '0');
        master_word_counter <= "0000";
        if master_readback = '1' then
          state <= ReadSPI0;
          master_readback_num <= std_logic_vector(unsigned('0' & readback_num) + 1);
          master_readback_reg <= readback_reg;
          master_word_counter <= x"F";
        else
          state <= ReadPage;
        end if;
      when ReadPage =>
        master_flash_go <= '1';
        state <= WaitFlash1;
      when WritePage =>
        master_flash_go <= '1';
        master_flash_command <= "010";
        state <= WaitFlash1;
        master_ram_write_i <= '0';
      when WaitFlash1 =>
        state <= WaitFlash2;
      when WaitFlash2 =>
        state <= WaitFlash2;
        if (flash_busy = '0' and master_flash_go = '0') then
          if master_readback = '1' then
            if master_readback_num = "00000" then
              state <= DisableFLASH1;
            else
              state <= ReadSPI0;
              master_flash_page <= std_logic_vector(unsigned(master_flash_page) + 1);
              --master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
            end if;
          else
            state <= ReadRAM;
          end if;
        end if;
        
      --read from registers and write to flash
      when ReadSPI0 =>
        master_ADDR_OUT <= master_readback_reg;
        master_READ_OUT <= '1';
        state <= ReadSPI1;
        if master_readback_num = "00000" then
          master_ram_data_i  <= x"00"; --stop
          state <= ReadSPI3;
        else
          master_ram_data_i  <= x"01"; --16bit version only
        end if;
        master_ram_write_i <= '1';
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
      when ReadSPI1 =>
        if LOC_READY_IN = '1' then
          state <= ReadSPI2;
          master_ram_data_i <= master_readback_reg;
          master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        end if;
      when ReadSPI2 =>
        master_ram_data_i <= LOC_DATA_IN(15 downto 8);
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        state <= ReadSPI3;
      when ReadSPI3 =>
        master_ram_data_i <= LOC_DATA_IN(7 downto 0);
        -- prepare for next cycle
        if (master_word_counter = x"E" or master_readback_num = "00000") then
          state <= WritePage;
        else
          state <= ReadSPI0;
        end if;
        if master_readback_num = "00000" then
          master_ram_write_i <= '0';
        else
          master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
          master_readback_num <= std_logic_vector(unsigned(master_readback_num) - 1);
          master_readback_reg <= std_logic_vector(unsigned(master_readback_reg) + 1);
        end if;
        
      --READ from flash and store to registers
      when ReadRAM =>
        state <= ReadRAM0;
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
      when ReadRAM0 =>
        state <= WaitRAM1;
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
      when WaitRAM1 =>
        if (ram_data_o = x"01") then --Version: 16 Bit
          state <= WaitRAM2;
          master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        elsif (ram_data_o = x"02" and DATA_BUS_WIDTH > 24) then --Version: 32 Bit
          state <= WaitRAM2_32;
          master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        else -- break
          master_DATA_OUT(15 downto 0) <= x"ff" & ram_data_o;
          state <= DisableFLASH1;
        end if;
      when WaitRAM2 =>
        master_ADDR_OUT <= ram_data_o;
        state <= WaitRAM3;
        --master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
      when WaitRAM2_32 =>
        master_ADDR_OUT <= ram_data_o;
        state <= WaitRAM3_32;
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
      when WaitRAM3 =>
        master_DATA_OUT(15 downto 8) <= ram_data_o;
        state <= WaitRAM4;
      when WaitRAM3_32 =>
        if (DATA_BUS_WIDTH > 31) then
          master_DATA_OUT(31 downto 0) <= ram_data_o & x"000000";
        end if;
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        state <= WaitRAM4_32;
      when WaitRAM4 =>
        master_DATA_OUT(15 downto 0) <= master_DATA_OUT(15 downto 8) & ram_data_o;
        state <= WriteSPI;
        master_WRITE_OUT <= '1';
      when WaitRAM4_32 =>
        if (DATA_BUS_WIDTH > 31) then
          master_DATA_OUT(31 downto 0) <= master_DATA_OUT(31 downto 24) & ram_data_o & x"0000";
        end if;
        --master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 2); -- padding bytes
        state <= WaitRAM5_32;
      when WaitRAM5_32 =>
        if (DATA_BUS_WIDTH > 31) then
          master_DATA_OUT(31 downto 0) <= master_DATA_OUT(31 downto 16) & ram_data_o & x"00";
        end if;
        --master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 2); -- padding bytes
        state <= WaitRAM6_32;
      when WaitRAM6_32 =>
        if (DATA_BUS_WIDTH > 31) then
          master_DATA_OUT(31 downto 0) <= master_DATA_OUT(31 downto 8) & ram_data_o;
        end if;
        state <= WriteSPI;
        master_WRITE_OUT <= '1';
      when WriteSPI =>
        -- prepare for next cycle
        if (master_word_counter = x"F") then
          state <= ReadPage;
          master_word_counter <= "0000";
          master_flash_page <= std_logic_vector(unsigned(master_flash_page) + 1);
        else
          state <= ReadRAM;
          master_word_counter <= std_logic_vector(unsigned(master_word_counter) + 1);
        end if;
      when DisableFLASH1 =>
        state <= DisableFLASH2;
        master_flash_command <= "101";
        master_flash_go <= '1';
      when DisableFLASH2 =>
        state <= DisableFLASH3;
        --master_flash_go <= '0';
      when DisableFLASH3 =>
        if (flash_busy = '0' and master_flash_go = '0') then
           state <= IDLE;
           master_flash_command <= "000";
        end if;
                             
    end case;
  end if;
end process;
    

end architecture;

	
	

